(* Time-stamp: <modified the 29/11/2013 (at 15:23) by Erwan Jahier> *)
(*-----------------------------------------------------------------------
** Copyright (C) - Verimag.
*)

type vars = (string * string) list

open Lv6MainArgs
open Soc
open SocExecValue

(* xxx avoidable code duplic with main.ml *)
let make argv =
  let opt = Lv6MainArgs.parse argv in
  let node = opt.main_node in
  Verbose.exe ~level:3 (fun () ->
    Gc.set { (Gc.get ()) with Gc.verbose = 0x01 }
  );
  if (opt.infiles = []) then (
    Lv6MainArgs.usage stderr opt;
    exit 1
  );
  let new_dft_pack = Filename.basename (Filename.chop_extension (List.hd opt.infiles)) in
  Ident.set_dft_pack_name new_dft_pack;

  let main_node = 
    if opt.main_node = "" then None else 
      Some (Ident.idref_of_string opt.main_node)
  in
  if opt.outfile <> "" then opt.oc <- open_out opt.outfile;
  let nsl = Compile.get_source_list opt opt.infiles in
  let lic_prg = Compile.doit opt nsl main_node in

  let first_file = List.hd opt.infiles in
  let nk = (Lic.node_key_of_idref (Ident.to_idref opt.main_node)) in
  let sk, soc_tbl = 
    if LicPrg.node_exists lic_prg nk then (
      Lic2soc.f lic_prg nk 
    ) else (
      print_string ("Error: cannot find node "^opt.main_node^" in "^
                       (String.concat "," opt.infiles)^".\n");
      flush stdout;
      exit 1
    )
  in
  let soc = try Soc.SocMap.find sk soc_tbl with Not_found -> assert false in
  let vntl_of_profile = List.map (fun (x,t) -> x,SocUtils.string_of_type_ref t) in
  let soc_inputs  = (SocExec.expand_profile false (fst soc.profile)) in
  let soc_outputs = (SocExec.expand_profile true (snd soc.profile)) in
  let (vntl_i:Data.vntl) = vntl_of_profile soc_inputs in
  let (vntl_o:Data.vntl) = vntl_of_profile soc_outputs in
  let oc = stdout in
(*    Lv6util.dump_entete oc;  *)
(*    RifIO.write_interface oc vntl_i vntl_o None None;  *)
(*    RifIO.flush oc;  *)

  let (to_soc_subst : SocExecValue.ctx -> Soc.var list -> Data.subst list) =
    fun ctx vl ->
(*       let sl = List.map (fun var -> fst var, SocExecValue.get_value ctx (Var var)) vl in *)
      let sl = SocExecValue.filter_top_subst ctx.s in
      let sl = List.flatten (List.map SocExec.expand_subst sl) in
  (* If the order ever matters, I could try the following.  :
     try List.map (fun v -> fst v,
     List.assoc (fst v) sl) vl with Not_found -> assert false
  *)
      sl
  in
  let (add_subst : Data.subst list -> SocExecValue.substs -> SocExecValue.substs) = 
    fun s ctx_s ->
      let s = SocExec.unexpand_profile s (fst soc.profile) in
      List.fold_left (fun acc (id,v) -> SocExecValue.sadd acc [id] v) ctx_s s
  in
  let ctx_ref = ref (SocExecValue.create_ctx soc_tbl soc) in
  let step sl_in =
    let ctx = { !ctx_ref with s = add_subst sl_in !ctx_ref.s } in
    let ctx = SocExec.do_step soc_tbl soc ctx in
    let sl_out = to_soc_subst ctx soc_outputs in
    ctx_ref := ctx;
(*     RifIO.write_outputs oc Util.my_string_of_float  vntl_o sl_out; *)
(*     RifIO.flush oc; *)
    sl_out
  in
  let step_dbg sl_in ectx cont =
    let cont2 ctx =
      let sl_out = to_soc_subst ctx soc_outputs in
      ctx_ref := ctx;
      cont sl_out ectx
    in
    ctx_ref := { !ctx_ref with s = add_subst sl_in !ctx_ref.s };
    SocExec.do_step_dbg soc_tbl soc ectx !ctx_ref cont2
  in 
  let (mems_in  : Data.subst list) = [] in (* XXX todo *)
  let (mems_out : Data.subst list) = [] in (* XXX todo *)
  vntl_i,vntl_o, (fun _ -> ()), step, step_dbg, mems_in, mems_out

